#include "stdafx.h"
#include "MyEditMisc.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/////////////////////////////////////////////////////////////////////////////
CMyEditTypeInfo::CMyEditTypeInfo()
{
        m_bLoaded        = false;
        m_bModified      = false;
        m_nMenuItemID    = 0;
        m_bDrawHilighted = true;
}

CMyEditTypeInfo& CMyEditTypeInfo::operator=(const CMyEditTypeInfo& src)
{
        // CPfxEditTypeInfo֡
        m_bDrawHilighted    = src.m_bDrawHilighted;
        m_bDrawRuler        = src.m_bDrawRuler;
        m_bDrawLineNumber   = src.m_bDrawLineNumber;
        m_bDrawLineBreak    = src.m_bDrawLineBreak;
        m_bDrawWhitespace   = src.m_bDrawWhitespace;
        m_bEnableDragDrop   = src.m_bEnableDragDrop;

        m_bExpandTabToSpaceChars                = src.m_bExpandTabToSpaceChars;
        m_bHomeKeyGotoFirstNonWhitespaceChar    = src.m_bHomeKeyGotoFirstNonWhitespaceChar;
        m_bSyncActiveLineBkColorWithTextBkColor = src.m_bSyncActiveLineBkColorWithTextBkColor;
        m_bSyncTextBkColorWithSystem            = src.m_bSyncTextBkColorWithSystem;
        
        m_nIndentSize       = src.m_nIndentSize;
        m_nTabSize          = src.m_nTabSize;
        m_nWrapStyle        = src.m_nWrapStyle;
        m_nWrapWidth        = src.m_nWrapWidth;

        // ɫ
        m_arrKeyClasses     = src.m_arrKeyClasses;
        m_arrMarkerPair     = src.m_arrMarkerPair;

        SetMatchCase(src.IsMatchCase());
        memcpy(m_arrDelimiters, src.m_arrDelimiters, sizeof(m_arrDelimiters));

        // CMyEditTypeInfo֡
        if (m_strTypeDescription != src.m_strTypeDescription)
        {
                m_strTypeDescription = src.m_strTypeDescription;
        }

        if (m_strTypeExtensions != src.m_strTypeExtensions)
        {
                m_strTypeExtensions = src.m_strTypeExtensions;
        }

        if (m_strTypeProfileName != src.m_strTypeProfileName)
        {
                m_strTypeProfileName = src.m_strTypeProfileName;
        }
        m_bLoaded   = src.m_bLoaded;
        m_bModified = src.m_bModified;

        return *this;
}

/////////////////////////////////////////////////////////////////////////////
IMPLEMENT_DYNAMIC(CMyEditTypeInfoList, CObject)
CMyEditTypeInfoList::CMyEditTypeInfoList(int nBlockSize)
{
        ASSERT(nBlockSize > 0);
        
        m_nCount        = 0;
        m_pNodeHead     = m_pNodeTail = m_pNodeFree = NULL;
        m_pBlocks       = NULL;
        m_nBlockSize    = nBlockSize;
}

void CMyEditTypeInfoList::RemoveAll()
{
        ASSERT_VALID(this);
        
        // destroy elements
        for (CNode* pNode = m_pNodeHead; pNode != NULL; pNode = pNode->pNext)
        {
#if (_MSC_VER <= 1200) // MSVC6.0
                DestructElements<CMyEditTypeInfo>(&pNode->data, 1);
#else
                pNode->data.~CMyEditTypeInfo();
#endif
        }
        
        m_nCount = 0;
        m_pNodeHead = m_pNodeTail = m_pNodeFree = NULL;
        m_pBlocks->FreeDataChain();
        m_pBlocks = NULL;
}

CMyEditTypeInfoList::~CMyEditTypeInfoList()
{
        RemoveAll();
        ASSERT(m_nCount == 0);
}

/////////////////////////////////////////////////////////////////////////////
// Node helpers
/*
* Implementation note: CNode's are stored in CPlex blocks and
*  chained together. Free blocks are maintained in a singly linked list
*  using the 'pNext' member of CNode with 'm_pNodeFree' as the head.
*  Used blocks are maintained in a doubly linked list using both 'pNext'
*  and 'pPrev' as links and 'm_pNodeHead' and 'm_pNodeTail'
*   as the head/tail.
*
* We never free a CPlex block unless the List is destroyed or RemoveAll()
*  is used - so the total number of CPlex blocks may grow large depending
*  on the maximum past size of the list.
*/

CMyEditTypeInfoList::CNode*
CMyEditTypeInfoList::NewNode(CMyEditTypeInfoList::CNode* pPrev, CMyEditTypeInfoList::CNode* pNext)
{
        if (m_pNodeFree == NULL)
        {
                // add another block
                CPlex* pNewBlock = CPlex::Create(m_pBlocks, m_nBlockSize, sizeof(CNode));
                
                // chain them into free list
                CNode* pNode = (CNode*) pNewBlock->data();

                // free in reverse order to make it easier to debug
                pNode += m_nBlockSize - 1;
                for (int i = m_nBlockSize-1; i >= 0; i--, pNode--)
                {
                        pNode->pNext = m_pNodeFree;
                        m_pNodeFree = pNode;
                }
        }
        ASSERT(m_pNodeFree != NULL);  // we must have something
        
        CMyEditTypeInfoList::CNode* pNode = m_pNodeFree;
        m_pNodeFree = m_pNodeFree->pNext;
        pNode->pPrev = pPrev;
        pNode->pNext = pNext;
        m_nCount++;
        ASSERT(m_nCount > 0);  // make sure we don't overflow
        
        
#if (_MSC_VER <= 1200) // MSVC6.0
        ConstructElements<CMyEditTypeInfo>(&pNode->data, 1);
#else
#pragma push_macro("new")
#undef new
        ::new((void*)(&pNode->data)) CMyEditTypeInfo;
#pragma pop_macro("new")
#endif
        
        return pNode;
}

void CMyEditTypeInfoList::FreeNode(CMyEditTypeInfoList::CNode* pNode)
{
#if (_MSC_VER <= 1200) // MSVC6.0
        DestructElements<CMyEditTypeInfo>(&pNode->data, 1);
#else
        pNode->data.~CMyEditTypeInfo();
#endif
        
        pNode->pNext = m_pNodeFree;
        m_pNodeFree = pNode;
        m_nCount--;
        ASSERT(m_nCount >= 0);  // make sure we don't underflow
        
        // if no more elements, cleanup completely
        if (m_nCount == 0)
        {
                RemoveAll();
        }
}

/////////////////////////////////////////////////////////////////////////////
POSITION CMyEditTypeInfoList::AddHead(const CMyEditTypeInfo* newElement)
{
        ASSERT_VALID(this);
        
        CNode* pNewNode = NewNode(NULL, m_pNodeHead);
        pNewNode->data = *newElement;
        if (m_pNodeHead != NULL)
        {
                m_pNodeHead->pPrev = pNewNode;
        }
        else
        {
                m_pNodeTail = pNewNode;
        }
        
        m_pNodeHead = pNewNode;
        return (POSITION) pNewNode;
}

POSITION CMyEditTypeInfoList::AddTail(const CMyEditTypeInfo* newElement)
{
        ASSERT_VALID(this);
        
        CNode* pNewNode = NewNode(m_pNodeTail, NULL);
        pNewNode->data = *newElement;
        if (m_pNodeTail != NULL)
        {
                m_pNodeTail->pNext = pNewNode;
        }
        else
        {
                m_pNodeHead = pNewNode;
        }
        
        m_pNodeTail = pNewNode;
        return (POSITION) pNewNode;
}

void CMyEditTypeInfoList::RemoveAt(POSITION position)
{
        ASSERT_VALID(this);
        
        CNode* pOldNode = (CNode*) position;
        ASSERT(AfxIsValidAddress(pOldNode, sizeof(CNode)));
        
        // remove pOldNode from list
        if (pOldNode == m_pNodeHead)
        {
                m_pNodeHead = pOldNode->pNext;
        }
        else
        {
                ASSERT(AfxIsValidAddress(pOldNode->pPrev, sizeof(CNode)));
                pOldNode->pPrev->pNext = pOldNode->pNext;
        }
        
        if (pOldNode == m_pNodeTail)
        {
                m_pNodeTail = pOldNode->pPrev;
        }
        else
        {
                ASSERT(AfxIsValidAddress(pOldNode->pNext, sizeof(CNode)));
                pOldNode->pNext->pPrev = pOldNode->pPrev;
        }
        FreeNode(pOldNode);
}


/////////////////////////////////////////////////////////////////////////////
// slow operations

POSITION CMyEditTypeInfoList::FindIndex(int nIndex) const
{
        ASSERT_VALID(this);
        
        if (nIndex >= m_nCount || nIndex < 0)
        {
                return NULL;  // went too far
        }
        
        CNode* pNode = m_pNodeHead;
        while (nIndex--)
        {
                ASSERT(AfxIsValidAddress(pNode, sizeof(CNode)));
                pNode = pNode->pNext;
        }
        return (POSITION) pNode;
}

POSITION CMyEditTypeInfoList::Find(CMyEditTypeInfo* searchValue, POSITION startAfter) const
{
        ASSERT_VALID(this);
        
        CNode* pNode = (CNode*) startAfter;
        if (pNode == NULL)
        {
                pNode = m_pNodeHead;  // start at head
        }
        else
        {
                ASSERT(AfxIsValidAddress(pNode, sizeof(CNode)));
                pNode = pNode->pNext;  // start after the one specified
        }
        
        for (; pNode != NULL; pNode = pNode->pNext)
        {
                if (&pNode->data == searchValue)
                {
                        return (POSITION) pNode;
                }
        }
        return NULL;
}


/////////////////////////////////////////////////////////////////////////////
// Diagnostics

#ifdef _DEBUG
void CMyEditTypeInfoList::Dump(CDumpContext& dc) const
{
        CObject::Dump(dc);
        
        dc << "with " << m_nCount << " elements";
        if (dc.GetDepth() > 0)
        {
                //POSITION pos = GetHeadPosition();
                //while (pos != NULL)
                //	dc << "\n\t" << GetNext(pos);
        }
        
        dc << "\n";
}

void CMyEditTypeInfoList::AssertValid() const
{
        CObject::AssertValid();
        
        if (m_nCount == 0)
        {
                // empty list
                ASSERT(m_pNodeHead == NULL);
                ASSERT(m_pNodeTail == NULL);
        }
        else
        {
                // non-empty list
                ASSERT(AfxIsValidAddress(m_pNodeHead, sizeof(CNode)));
                ASSERT(AfxIsValidAddress(m_pNodeTail, sizeof(CNode)));
        }
}
#endif //_DEBUG
